/* Functors used as callbacks by the event-handler.
   Copyright (C) 2006 Carl Fredrik Hammar (hammy.lite@gmail.com)

   This file is part of GP2X Sand.

   GP2X Sand is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   GP2X Sand is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details. 

   You should have received a copy of the GNU General Public License along
   with GP2X Sand; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */

#include "action.h"

#include <assert.h>

/* This module is centered around actions, simple functors that handle
   certain types of events.  And also contains some convenient utilities
   that are useful for creating actions.  As actions are used to handle
   events they are used in conjunction with the eventhandler module. */

/* Forward declarations. */
void
action_plain_function (void *action_ptr, SDL_Event *event_ptr);

/* Initialize the action pointed to by ACTION_PTR to use FUNCTION.
   ACTION_PTR mustn't be null.  If FUNCTION is null this action does
   nothing on invocation. */
void
action_init (action_t *action_ptr, action_function_t function)
{
  assert (action_ptr != 0);
  assert (function != 0);

  action_ptr->function = function;
}

/* Invokes the action ACTION_PTR points to.  The invocation calls the
   function assigned to the action with ACTION_PTR and EVENT_PTR as
   parameters. If ACTION_PTR is null or the actions assigned function is
   null, nothing is done.  EVENT_PTR mustn't be null. */
void
action_invoke (action_t *action_ptr, SDL_Event *event_ptr)
{
  if (action_ptr != 0 && action_ptr->function != 0)
    action_ptr->function (action_ptr, event_ptr);
}

/* Initialize the plain action pointed to by PLAIN_ACTION_PTR to call
   FUNCTION on invokation when the invoking event is a trigger with state
   TRIGGER_STATE.  PLAIN_ACTION_PTR mustn't be null.  If FUNCTION is null
   invoking this action does nothing. */  
void
action_plain_init (action_plain_t *action_plain_ptr,
		   action_plain_function_t function,
		   trigger_state_t trigger_state)
{
  assert (action_plain_ptr != 0);
  assert (function != 0);
  
  action_init (&action_plain_ptr->action, action_plain_function);
  action_plain_ptr->function = function;
  action_plain_ptr->trigger_state = trigger_state;
}

/* Converts ACTION_PTR to a action_plain_t pointer and fetches it's
   function and trigger state and calls the function if the trigger state
   of EVENT_PTR matches that of the plain action. */
void
action_plain_function (void *action_ptr, SDL_Event *event_ptr)
{
  action_plain_t *action_plain_ptr = action_ptr;
  
  if (action_plain_ptr != 0
      && action_plain_ptr->function != 0
      && (event_trigger_state (event_ptr)
	  == action_plain_ptr->trigger_state))
    action_plain_ptr->function ();
}
  
/* Return the trigger state of the event pointed to by EVENT_PTR.  Returns
   TRIGGER_DOWN and TRIGGER_UP on respective button-like events,
   TRIGGER_DOWN on SDL_QUIT, TRIGGER_UNKNOWN for any other type of
   event. EVENT_PTR mustn't be null. */
trigger_state_t
event_trigger_state (SDL_Event *event_ptr)
{
  assert (event_ptr != 0);
  
  switch (event_ptr->type)
    {
    case SDL_QUIT:
    case SDL_JOYBUTTONDOWN:
    case SDL_KEYDOWN:
    case SDL_MOUSEBUTTONDOWN:
      return TRIGGER_DOWN;
      
    case SDL_JOYBUTTONUP:
    case SDL_KEYUP:
    case SDL_MOUSEBUTTONUP:
      return TRIGGER_UP;

    default:
      return TRIGGER_UNKNOWN;
    }
}

