/* Handle the stick of the GP2X.
   Copyright (C) 2006 Carl Fredrik Hammar (hammy.lite@gmail.com)

   This file is part of GP2X Sand.

   GP2X Sand is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   GP2X Sand is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details. 

   You should have received a copy of the GNU General Public License along
   with GP2X Sand; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */

#include "gp2xstick.h"

#include <stdbool.h>
#include <assert.h>

#include "eventhandler.h"
#include "gp2xbuttons.h"
#include "pen.h"
#include "utils.h"

/** This module handles the gp2x's stick and maps it to move the pen
    around. **/

/* Flags for direction. */
enum {
  CENTER    = 0,
  UP        = 1,
  DOWN      = 1 << 1,
  LEFT      = 1 << 2,
  RIGHT     = 1 << 3,
  UPLEFT    = 1 << 4,
  UPRIGHT   = 1 << 5,
  DOWNLEFT  = 1 << 6,
  DOWNRIGHT = 1 << 7
};

int stick_direction;
int cursor_speed;
int cursor_velocity_x;
int cursor_velocity_y;


/** Forward declarations. **/
void
stick_actions_init ();
void
stick_action_motion_function (void *action, SDL_Event *event);
extern action_t *const stick_action_motion_ptr;

/* Initialize the stick.  This remaps gp2x stick events in the
   event-handler. */
void
stick_init_default ()
{
  stick_direction = 0;
  cursor_speed = 3;
  cursor_velocity_x = 0;
  cursor_velocity_y = 0;

  stick_actions_init ();
  
  int stick_buttons[] = { GP2X_BUTTON_UP,
			  GP2X_BUTTON_DOWN,
			  GP2X_BUTTON_LEFT,
			  GP2X_BUTTON_RIGHT,
			  GP2X_BUTTON_UPLEFT,
			  GP2X_BUTTON_UPRIGHT,
			  GP2X_BUTTON_DOWNLEFT,
			  GP2X_BUTTON_DOWNRIGHT };
 
  for (int i = 0; i < LENGTH_OF (stick_buttons); i++)
    eventhandler_joybutton_map (stick_buttons[i],
				stick_action_motion_ptr); 
}

/* Examines the current state of the stick and moves the pen
   accordingly.  Call this regularly. */
void
stick_pen_move ()
{
  // Determine vector from stick direction.
  if (stick_direction == 0)
    {
      cursor_velocity_x = 0;
      cursor_velocity_y = 0;
    }

  else if (stick_direction & UP)
    {
      cursor_velocity_x = 0;
      cursor_velocity_y = -cursor_speed;
    }

  else if (stick_direction & DOWN)
    {
      cursor_velocity_x = 0;
      cursor_velocity_y = cursor_speed;
    }

  else if (stick_direction & LEFT)
    {
      cursor_velocity_x = -cursor_speed;
      cursor_velocity_y = 0;
    }

  else if (stick_direction & RIGHT)
    {
      cursor_velocity_x = cursor_speed;
      cursor_velocity_y = 0;
    }

  else if (stick_direction & UPLEFT)
    {
      cursor_velocity_x = -cursor_speed;
      cursor_velocity_y = -cursor_speed;
    }

  else if (stick_direction & UPRIGHT)
    {
      cursor_velocity_x = cursor_speed;
      cursor_velocity_y = -cursor_speed;
    }

  else if (stick_direction & DOWNLEFT)
    {
      cursor_velocity_x = -cursor_speed;
      cursor_velocity_y = cursor_speed;
    }

  else if (stick_direction & DOWNRIGHT)
    {
      cursor_velocity_x = cursor_speed;
      cursor_velocity_y = cursor_speed;
    }

  // Cursor movement by joystick.
  if (stick_direction != CENTER)
    {
      pen_move (cursor_velocity_x,  cursor_velocity_y);
      SDL_WarpMouse (pen.x, pen.y);
    }
}

/** Stick actions. **/

action_t stick_action_motion;
action_t *const stick_action_motion_ptr = &stick_action_motion;

void
stick_actions_init ()
{
  stick_action_motion.function = stick_action_motion_function;
}

void
stick_action_motion_function (void *action, SDL_Event *event)
{
  int direction = 0;

  switch(event->jbutton.button)
    {
    case GP2X_BUTTON_UP:
      direction = UP;
      break;
		  
    case GP2X_BUTTON_DOWN:
      direction = DOWN;
      break;
		  
    case GP2X_BUTTON_LEFT:
      direction = LEFT;
      break;
	  
    case GP2X_BUTTON_RIGHT:
      direction = RIGHT;
      break;
	  
    case GP2X_BUTTON_UPLEFT:
      direction = UPLEFT;
      break;
	  
    case GP2X_BUTTON_UPRIGHT:
      direction = UPRIGHT;
      break;
	  
    case GP2X_BUTTON_DOWNLEFT:
      direction = DOWNLEFT;
      break;
	  
    case GP2X_BUTTON_DOWNRIGHT:
      direction = DOWNRIGHT;
      break;

    default:
      break;
    }
      
  if (event_trigger_state (event) == TRIGGER_DOWN)
    stick_direction |= direction;

  else if (event_trigger_state (event) == TRIGGER_UP)
    stick_direction &= ~direction;
}
