/* Pen for drawing particles to the sandbox.
   Copyright (C) 2006 Russell Turner,
                      Thomas Ren Sidor,
                      Carl Fredrik Hammar (hammy.lite@gmail.com)

   This file is part of GP2X Sand.

   GP2X Sand is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   GP2X Sand is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details. 

   You should have received a copy of the GNU General Public License along
   with GP2X Sand; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */

#include "pen.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>

#include "action.h"
#include "globals.h"
#include "sandbox.h"
#include "utils.h"

/** The pen draws a filled circle of new particles centered on the pen's
    current position and as it moves around.
    It has the following attributes:

    Position:      The x and y coordinates of it's current position.
    Radius:        The radius of the circle to draw.
    Paint element: The element to fill the circle with.
    Draw flag:     Only draws if this flag is set. **/

/* The actual pen. */
pen_t pen;

/** Forward declarations. **/
void
draw_particles (int xpos, int ypos, int radius, element_t *element_ptr);
void
draw_line (int newx, int newy, int oldx, int oldy);
void
pen_actions_init ();
void
pen_paint_element_cycle_forward ();
void
pen_paint_element_cycle_backward ();
void
pen_action_draw_function (void *action_ptr,
			  SDL_Event *event_ptr);

/* Initialize the pen with default values. */
void
pen_init_default ()
{
  pen.x = 0;
  pen.y = 0;
  pen.radius = 2;
  pen.paint_element_ptr = element_wall_ptr;
  pen.draw_flag = false;

  pen_actions_init ();
}

/* Draw a filled circle of particles centered at the pen's position with
   the radius as stored in the pen.  The element painted is the pen's
   paint_element. */ 
void
pen_draw ()
{
  pen_move (0, 0);
}

/* Move the pen horizontally and vertically, DX and DY pixels
   respectively.  Drawing a circle of particles at each pixel moved
   through.  The pen may not leave the sandbox, so if (X, Y) is outside
   the sandbox the new position of the pen is the one closest to (X, Y)
   within the sandbox. */
void
pen_move (int dx, int dy)
{
  pen_warp (pen.x + dx, pen.y + dy);
}

/* Move the pen horizontally and vertically, DX and DY pixels
   respectively.  Drawing a circle of particles at each pixel moved
   through.  The pen may not leave the sandbox, so if (X, Y) is outside
   the sandbox the new position of the pen is the one closest to (X, Y)
   within the sandbox. */
void
pen_warp (int x, int y)
{
  if (pen.draw_flag)
    draw_line (pen.x, pen.y, x, y);

  pen.x = clip_to_interval (0, SCREEN_WIDTH - 1, x);
  pen.y = clip_to_interval (0, SCREEN_HEIGHT - 1, y);
}

/* Halve the radius of the circle drawn by the pen. */
void
pen_radius_halve ()
{
  pen.radius /= 2;
  if (pen.radius < 1)
    pen.radius = 1;
}

/* Double the radius of the circle drawn by the pen. */
void
pen_radius_double ()
{
  pen.radius *= 2;
  if (pen.radius > SCREEN_HEIGHT / 2)
    pen.radius = SCREEN_HEIGHT / 2;
}

/* Set the element the pen draws to PAINT_ELEMENT. */
void
pen_paint_element_set (element_t *paint_element_ptr)
{
  assert (paint_element_ptr != 0);
  
  pen.paint_element_ptr = paint_element_ptr;
}

/* Set the flag that enables/disables drawing by the pen. */
void
pen_draw_flag_set (bool draw_flag)
{
  pen.draw_flag = draw_flag;
}
  
void
pen_paint_element_cycle_forward ()
{
  if (pen.paint_element_ptr == sandbox_element_last ())
    pen.paint_element_ptr = sandbox_element_first ();

  else
    pen.paint_element_ptr = sandbox_element_next (pen.paint_element_ptr);
  
}

void
pen_paint_element_cycle_backward ()
{
  if (pen.paint_element_ptr == sandbox_element_first ())
    pen.paint_element_ptr = sandbox_element_last ();

  else
    pen.paint_element_ptr = sandbox_element_prev (pen.paint_element_ptr);
}

// Drawing a filled circle at a given position with a given radius and a
// given partice type
void
draw_particles (int xpos, int ypos, int radius, element_t *element_ptr)
{
  for (int x = ((xpos - radius - 1) < 0) ? 0 : (xpos - radius - 1);
       x <= xpos + radius && x < SCREEN_WIDTH;
       x++)
    for (int y = ((ypos - radius - 1) < 0) ? 0 : (ypos - radius - 1);
	 y <= ypos + radius && y < SCREEN_HEIGHT;
	 y++)
      {
	if ((x - xpos) * (x - xpos) + (y - ypos) * (y - ypos)
	    <= radius * radius)
	  sandbox_particle_init (vcons( x, y), element_ptr);
      };
}

// Drawing a line
void
draw_line (int newx, int newy, int oldx, int oldy)
{
  if (newx == oldx && newy == oldy)
    {
      draw_particles (newx, newy, pen.radius, pen.paint_element_ptr);
    }
  else
    {
      float step = 1.0f / (abs(newx - oldx) > abs(newy - oldy)
			   ? abs(newx - oldx) : abs(newy - oldy));
      for (float a = 0; a < 1; a += step)
	draw_particles (lrint(a * newx + (1 - a) * oldx),
			lrint(a * newy + (1 - a) * oldy),
			pen.radius, pen.paint_element_ptr); 
    }
}

/** Pen actions **/

/* Sets the pen to draw when trigger is down and resets it when trigger is
   up. */
action_t pen_action_draw;

/* The following two actions simply wrap pen_radius_double() and
   pen_radius_halve() respectively. */
action_plain_t pen_action_size_increase;
action_plain_t pen_action_size_decrease;

/* The following two actions cycles through the available elements setting
   the element to pain by the pen to the next and previous element
   respectively. */
action_plain_t pen_action_paint_element_cycle_forward;
action_plain_t pen_action_paint_element_cycle_backward;

/* Initialize the pointers to the actions. */
action_t *const pen_action_draw_ptr = &pen_action_draw;
action_t *const pen_action_size_increase_ptr =
  (action_t*) &pen_action_size_increase;
action_t *const pen_action_size_decrease_ptr =
  (action_t*) &pen_action_size_decrease;
action_t *const pen_action_paint_element_cycle_forward_ptr =
  (action_t*) &pen_action_paint_element_cycle_forward;
action_t *const pen_action_paint_element_cycle_backward_ptr =
  (action_t*) &pen_action_paint_element_cycle_backward;

void
pen_actions_init ()
{
  action_init (&pen_action_draw, pen_action_draw_function);

  action_plain_init (&pen_action_size_increase, pen_radius_double,
		     TRIGGER_DOWN);
  action_plain_init (&pen_action_size_decrease, pen_radius_halve,
		     TRIGGER_DOWN);
  
  action_plain_init (&pen_action_paint_element_cycle_forward,
		     pen_paint_element_cycle_forward,
		     TRIGGER_DOWN);
  
  action_plain_init (&pen_action_paint_element_cycle_backward,
		     pen_paint_element_cycle_backward,
		     TRIGGER_DOWN);
}

void
pen_action_draw_function (void *action_ptr, SDL_Event *event_ptr)
{
  if (event_trigger_state (event_ptr) == TRIGGER_DOWN)
    pen_draw_flag_set (true);

  else if (event_trigger_state (event_ptr) == TRIGGER_UP)
    pen_draw_flag_set (false);
}
